/*
 * IBM Confidential
 * OCO Source Materials
 * 5724-U98, 5655-V48, 5724-U50, 5655-V46
 * Copyright IBM Corp. 2013
 * The source code for this program is not published or other -
 * wise divested of its trade secrets, irrespective of what has
 * been deposited with the U.S. Copyright Office.
 */
header {
	package com.ibm.nex.parser.oim.distributed.columnexpression;
} 
{ 
import java.text.DateFormat;
import java.text.SimpleDateFormat;
import java.util.Calendar;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

@SuppressWarnings("all") }
class DistributedColumnExpressionParser extends Parser;

options {
	buildAST = true;
	k = 4;
	defaultErrorHandler=false;
}

{
	/**
	 * Invalid characters for Optim name
	 */
	protected static final String INVALID_OPTIM_NAME_CHARACTERS = "%~`!^&*()+=|{}[]:;\"\'<>,.?/\\";

	/**
	 * Invalid characters for DLL name
	 */
	protected static final String INVALID_DLL_NAME_CHARACTERS = "*()|:\"<>,.?/\\";

	/**
	 * Indicates the flavor is not determined
	 */
	public static final int unknownFlavor = 0;

	/**
	 * Indicates the string expression should be processed
	 */
	public static final int stringFlavor = 1 << 1;

	/**
	 * Indicates the numeric expression should be processed
	 */
	public static final int numericFlavor = 1 << 2;

	/**
	 * Indicates the binary expression should be processed
	 */
	public static final int binaryFlavor = 1 << 3;

	/**
	 * Indicates the date/time expression should be processed
	 */
	public static final int datetimeFlavor = 1 << 4;

	/**
	 * Indicates the boolean expression should be processed
	 */
	public static final int booleanFlavor = 1 << 5;

	/**
	 * The number of operators appeared
	 */	
	protected int numOperators = 0;

	/**
	 * The flavor of current expression
	 */
	protected int flavor = unknownFlavor;

	/**
	 * The column's length
	 */
	int length = Integer.MAX_VALUE;

	/**
	 * The column's precision
	 */
	int precision = Integer.MAX_VALUE;

	/**
	 * The column's scale
	 */
	int scale = Integer.MAX_VALUE;

	/**
	 * Indicates the scale of numeric column
	 */
	boolean nullable = true;

	/**
	 * Indicates if PROP() function is being processed
	 */
	protected int inPropFunc = 0;

	/**
	 * Determines if the token is a valid optim identifier.
	 * @param token a token
	 * @returns true if valid
	 */
	protected boolean isValidOptimIndentifierToken(Token token){
		boolean isValidToken = true;		
		String tokenText = token.getText();
		// Valid values are letters, '@', '#', '$', and '_'.  Digits are valid except as the first character.
		for(int charIndex=0; charIndex<tokenText.length(); charIndex++){
			char tokenCharacter = tokenText.charAt(charIndex);
			if (Character.isLetter(tokenCharacter)){
			    continue;
			}
			if (charIndex > 0) {
			    if (Character.isDigit(tokenCharacter)){
			        continue;
			    }
			    if ("@#$_".indexOf(tokenCharacter) >= 0) {
			        continue;
			    }
			}
			int intValue = (int) tokenCharacter;
			if (intValue > 0x00ff) {
				continue;
			}
			return false;
		}
		return true;
	}

	/**
	 * Determines if the token is a valid unsigned integer.
	 * @param token a token
	 * @returns true if valid
	 */
	protected boolean isValidUnsignedInteger(Token token){
		return isValidUnsignedInteger(token.getText());
	}
	
	protected boolean isValidUnsignedInteger(String string){
		return between(string, 0, Integer.MAX_VALUE);
	}
	
	/**
	 * Determines if the token is a valid signed integer.
	 * @param token a token
	 * @returns true if valid
	 */
	protected boolean isValidInteger(Token token){
		return isValidInteger(token.getText());
	}
	
	protected boolean isValidInteger(String string){
		return between(string, Integer.MIN_VALUE, Integer.MAX_VALUE);
	}
	
	/**
	 * Determines if the token is in a valid integer range.
	 * @param token a token
	 * @param low a lowest value
	 * @param high a highest value
	 * @returns true if the token represents an integer between low and high
	 */
	protected boolean between(Token token, long low, long high) {
		return between(token.getText(), low, high);
	}

	protected boolean between(String string, long low, long high) {
		if (string == null || string.isEmpty()) {
			return false;
		}
		try {
			long value = Long.parseLong(string);
			return low <= value && value <= high;
		}
		catch (NumberFormatException e) {
			return false;
		}
	}

	/**
	 * Determines if the token is a valid optim name.
	 * @param token a token
	 * @returns true if valid
	 */
	protected boolean isValidOptimNameToken(Token token){
		if (token == null || token.getText() == null || token.getText().isEmpty()) {
			return false;
		}
		String tokenText = token.getText();
		return !containsInvalidChars(tokenText, INVALID_OPTIM_NAME_CHARACTERS);
	}
	
	/**
	 * Determines if the token is a valid dll name.
	 * @param dllname a dll name
	 * @returns true if valid
	 */
	protected boolean isValidDllName(String dllname){
		if (dllname == null || dllname.isEmpty() || dllname.length() > 12) {
			return false;
		}
		return !containsInvalidChars(dllname, INVALID_DLL_NAME_CHARACTERS);
	}
	
	/**
	 * Determines if the token is a valid object identifier.
	 * @param token a token
	 * @param length a maximum string length
	 * @returns true if valid
	 */
	protected boolean isValidObjectIdentifier(Token token, int length){
		if (token == null) {
			return false;
		}
		String tokenText = token.getText();
		if (tokenText == null || tokenText.isEmpty() || tokenText.length() > length) {
			return false;
		}
		return isValidOptimIndentifierToken(token);
	}
	
	/**
	 * Determines if the token is a valid decimal string.
	 * @param token a token
	 * @returns true if valid
	 */
	protected boolean isValidDecimalString(Token token){
		return isValidDecimalString(token.getText());
	}

	protected boolean isValidDecimalString(String string){
		if (string == null || string.isEmpty()){
			return false;
		}
		string = string.toLowerCase();
		for(int charIndex=0; charIndex<string.length(); charIndex++){
			if (!Character.isDigit(string.charAt(charIndex))){
				return false;
			}
		}
		return true;		
	}
	
	/**
	 * Determines if the token is a valid hex string.
	 * @param token a token
	 * @returns true if valid
	 */
	protected boolean isValidHexString(Token token){
		return isValidHexString(token.getText());
	}

	protected boolean isValidHexString(String string){
		if (string == null || string.isEmpty()){
			return false;
		}
		string = string.toLowerCase();
		final String HEX_CHARS = "abcdef";
		for(int charIndex=0; charIndex<string.length(); charIndex++){
			if (!Character.isDigit(string.charAt(charIndex)) && !HEX_CHARS.contains(string.substring(charIndex, charIndex+1))){
				return false;
			}
		}
		return true;		
	}
	
	/**
	 * Determines if the string is a valid datetime string.
	 * @param string a string
	 * @returns true if valid
	 */
    private boolean isValidDateValueString(String string) {
        String regex =  "(\\d+)SEP(\\d+)SEP(\\d+)( +(\\d+)(SEP(\\d+)(SEP(\\d+)(\\.(\\d+))?)?)?)?".replace("SEP", "[/\\.\\-:]");
        Matcher matcher = Pattern.compile("^"+regex+"$").matcher(string.trim());
        if (!matcher.find()) {
            return false;
        }
        SimpleDateFormat sdf = (SimpleDateFormat)SimpleDateFormat.getDateInstance(DateFormat.SHORT);
        String pattern = sdf.toPattern().replaceAll("[-\\. ]", "/").replaceAll("[^yMd/]", "").replace("//", "/").replaceAll("/$", "")
                       . replaceAll("y+","y").replaceAll("M+","M").replaceAll("d+", "d");
        String year = matcher.group(pattern.indexOf('y') /2 + 1);
        String month = matcher.group(pattern.indexOf('M') /2 + 1);
        String date = matcher.group(pattern.indexOf('d') /2 + 1);
        String hour = matcher.group(5);  
        String minute = matcher.group(7);  
        String second = matcher.group(9);  
        String fraction = matcher.group(11);
        if (year == null || month == null || date == null) {
            return false;
        }
        Calendar cal = Calendar.getInstance();
        cal.set(Integer.valueOf(year), Integer.valueOf(month) - 1, Integer.valueOf(date));
        cal.setLenient(false);
        try {
            cal.getTime();
        } catch(Exception e) {
            return false;
        }
        if (hour != null && Integer.valueOf(hour) > 23 
                || minute != null && Integer.valueOf(minute) > 59 
                || second != null && Integer.valueOf(second) > 59 
                || fraction != null && fraction.length() > 6) {
            return false;
        }
        return true;
    }
    
	/**
	 * Handles cases where a keyword may be in place of a generic optim token.
	 * possibleTokenTypes: Other possible token types.  GENERICOPTIMTOKEN is assumed to be a valid type.
	 * @param possibleTokenTypes an array of token types able to appear
	 * @throws TokenStreamException anything that goes wrong while generating a stream of tokens
	 * @throws RecognitionException
	 */ 
	protected void handlePossibleGenericOptimToken(int [] possibleTokenTypes) throws TokenStreamException, RecognitionException {
		handlePossibleGenericOptimToken(1, possibleTokenTypes);
	}

	protected void handlePossibleGenericOptimToken(int tokenPosition, int [] possibleTokenTypes) throws TokenStreamException, RecognitionException {
		int nextTokenType = LA(tokenPosition);
		// If the token is already one of the possible token types, then do nothing else.
		if (possibleTokenTypes != null){
			for(int tokenType : possibleTokenTypes){
				if (nextTokenType == tokenType){
					return;
				}
			}
		}
		// If the token is not a GENERICOPTIMTOKEN, then check to see if it is a keyword that could be changed into a GENERICOPTIMTOKEN. 
		if (nextTokenType != GENERICOPTIMTOKEN) {
			Token token = LT(tokenPosition);
			if (isValidOptimNameToken(token)) {
				token.setType(GENERICOPTIMTOKEN);
			} else {
				throw new RecognitionException("Invalid value for genericOptimToken rule.", getFilename(), token.getLine(), token.getColumn());				
			}
		} else {
			Token token = LT(tokenPosition);
			String tokenText = token.getText();
			if (tokenText == null || tokenText.isEmpty() || tokenText.compareTo("\"\"") == 0) {
				throw new RecognitionException("Invalid value for genericOptimToken rule.", getFilename(), token.getLine(), token.getColumn());				
			}
		}
	}

	protected void handlePossibleGenericOptimToken(int startTokenPosition, int endTokenPosition, int [] possibleTokenTypes) throws TokenStreamException, RecognitionException {
		int tokenPosition;
		for (tokenPosition = startTokenPosition; tokenPosition <= endTokenPosition; tokenPosition++) {
			handlePossibleGenericOptimToken(tokenPosition, possibleTokenTypes);
		}
	}
	
	/**
	 * Handles look-ahead for a multi-part name with variable number of name parts.
	 * This method is invoked for both identifiers and names, but since name is the
	 * least restrictive, we'll check for name validity and let the actual rule which
	 * has called this method validate each part.
	 * @throws TokenStreamException anything that goes wrong while generating a stream of tokens
	 * @throws RecognitionException
	 */
	protected void handleMultiPartNameLookAhead() throws TokenStreamException, RecognitionException {
		int lookAheadTokenType;
    	handlePossibleGenericOptimToken(1, new int[] {DOUBLEQUOTEDSTRING});
        if (LA(2) == DOT) {
	    	handlePossibleGenericOptimToken(3, new int[] {DOUBLEQUOTEDSTRING});
        }
        if (LA(4) == DOT) {
	    	handlePossibleGenericOptimToken(5, new int[] {DOUBLEQUOTEDSTRING});
        }
	}

	/**
	 * Handles look-ahead for the flags following '=' character in TRANS functions.
	 * @throws TokenStreamException anything that goes wrong while generating a stream of tokens
	 * @throws RecognitionException
	 */
	protected void handleTransEmlLookAhead() throws TokenStreamException, RecognitionException {
		int possibleTokens[] = { DOUBLEQUOTEDSTRING, COMMA, SINGLEQUOTE, RIGHTPAREN };
		for (int tokenIndex = 1; tokenIndex <= 4; tokenIndex++) {
			if (LA(tokenIndex) == EOF) {
				break;
			}
			if (LA(tokenIndex) == COMMA) {
				continue;
			}
			if (LA(tokenIndex + 1) == EQUALS) {
				break;
			}
			handlePossibleGenericOptimToken(tokenIndex, possibleTokens);
		}
	}

	/**
	 * Handles look-ahead for the TRANS COL functions.
	 * @throws TokenStreamException anything that goes wrong while generating a stream of tokens
	 * @throws RecognitionException
	 */
	protected void handleTransColLookAhead() throws TokenStreamException, RecognitionException {
		int tokenIndex = 1;
		if (LA(tokenIndex) == COMMA) {
			tokenIndex++;
		}
		if (LA(tokenIndex) == LITERAL_SOURCE && LA(tokenIndex + 1) == EQUALS) {
			handlePossibleGenericOptimToken(tokenIndex + 2, new int[] { DOUBLEQUOTEDSTRING });
		}
	}

	/**
	 * Handles look-ahead for the oracle sequence.
	 * @throws TokenStreamException anything that goes wrong while generating a stream of tokens
	 * @throws RecognitionException 
	 */
	protected void handleOracleSequenceLookAhead() throws TokenStreamException, RecognitionException {
		if (possibleOracleSequence()) {
			handlePossibleGenericOptimToken(1,null);
			handlePossibleGenericOptimToken(3,null);
		}
	}

	/**
	 * Determines if the token contains all valid flags
	 * @param token a token
	 * @param flags a valid flags
	 * @returns true if valid
	 */
	protected boolean isValidFlags(Token token, String flags){
		return isValidFlags(token.getText(), flags);
	}

	protected boolean isValidFlags(String string, String flags){
		return countValidChars(string, flags) == string.length();
	}

	/**
	 * Determines if the token cotains conflict flags
	 * @param token a token
	 * @param flags characters conflicting with each others
	 * @returns true if conflict
	 */
	protected boolean containsConflictFlags(Token token, String flags){
		String tokenText = token.getText().toUpperCase();
		return countValidChars(flags, tokenText) > 1;
	}

	/**
	 * Counts the valid characters in the given string
	 * @param string a string
	 * @param validChars a string of valid characters
	 * @returns the count of valid characters
	 */
	protected int countValidChars(String string, String validChars){
		int count = 0;
		string = string.toUpperCase();
		validChars = validChars.toUpperCase();
		for(int charIndex=0; charIndex<string.length(); charIndex++){
			char ch = string.charAt(charIndex);
			if (validChars.indexOf(ch) >= 0){
				count++;
			}
		}		
		return count;
	}

	/**
	 * Determine if the token contains invalid characters
	 * @param string a string
	 * @param invalidChars a string of invalid characters
	 * @returns true if the string contains a invalid character
	 */
	protected boolean containsInvalidChars(String string, String invalidChars){
		string = string.toUpperCase();
		invalidChars = invalidChars.toUpperCase();
		for(int charIndex=0; charIndex<string.length(); charIndex++){
			char ch = string.charAt(charIndex);
			if (invalidChars.indexOf(ch) >= 0){
				return true;
			}
		}		
		return false;
	}

	/**
	 * Determines if the token contains a valid aging amount
	 * @param token a token
	 * @param plus indicates whether the plus sign is preceeded
	 * @param minus indicates whether the minus sign is preceeded
	 * @returns true if the token represents a valid aging amount
	 */
	protected boolean isValidAgingAmount(Token token, boolean plus, boolean minus) {
		String tokenText = token.getText().toUpperCase();
		if (tokenText.matches("^-?\\d+[YMWDR]$")) {
			int amount = Integer.valueOf(tokenText.substring(0, tokenText.length()-1));
			if (minus) {
				amount *= -1;
			}
		    char lastChar = tokenText.charAt(tokenText.length()-1);
			switch(lastChar) {
				case 'Y': 
					return (plus ? 0 : -2500) <= amount && amount <= 1581
						|| !plus && 1582 <= amount && amount <= 3999;
				case 'M': 
				case 'W':
					return (plus ? 0 : -30000) <= amount && amount <= 30000;
				case 'D':
					return (plus ? 0 : -99999) <= amount && amount <= 99999;
				case 'R':
					return ((plus ? 1 : -30000) <= amount && amount <= 30000) && amount != 0;
			}
		}
		return false;
	}
	
	/**
	 * Determines if the string contains a valid date format
	 * @param string a string
	 * @returns true if valid
	 */
	protected boolean isValidDateFormatString(String string) {
		return string.toUpperCase().matches("^(YYYY|CCYY|YY|MONTH|MMM|MM|M|DDD|DD|D|HH|MI|SS|FFFFFF|FFFFF|FFFF|FFF|FF|F|[^A-Z])+$");
	}
	
	/**
	 * Detects whitespaces between tokens
	 * @param tokenIndexStart the start index of tokens
	 * @param tokenIndexEnd the end index of tokens
	 * @returns true if there exists whitespaces between the given token range
	 * @throws TokenStreamException anything that goes wrong while generating a stream of tokens 
	 */
	protected boolean existsWhitespace(int tokenIndexStart, int tokenIndexEnd) throws TokenStreamException {
		for (int tokenIndex=tokenIndexStart; tokenIndex < tokenIndexEnd; tokenIndex++) {
			if (LT(tokenIndex).getColumn() + LT(tokenIndex).getText().length() != LT(tokenIndex+1).getColumn()) {
				return true;
			}
		}
		return false;
	}

	/**
	 * Counts clinging tokens (max 4)
	 * @throws TokenStreamException anything that goes wrong while generating a stream of tokens 
	 */
	protected int countClingingTokens() throws TokenStreamException {
		int tokenIndex;
		for (tokenIndex = 1; tokenIndex < 4; tokenIndex++) {
			if (LA(tokenIndex+1) == EOF || LT(tokenIndex).getColumn() + LT(tokenIndex).getText().length() != LT(tokenIndex+1).getColumn()) {
				break;
			}
		}
		return tokenIndex;
	}

	/**
	 * Determines if a oracle sequence appears in the token stream.
	 * @throws TokenStreamException anything that goes wrong while generating a stream of tokens 
	 */
	protected boolean possibleOracleSequence() throws TokenStreamException {
		return LA(5) == LITERAL_NEXTVAL && LA(2) == DOT && LA(4) == DOT && !existsWhitespace(1,5); 
	}

	/**
	 * Determines if a numeric constant appears in the token stream.
	 * @throws TokenStreamException anything that goes wrong while generating a stream of tokens 
	 */
	protected boolean possibleNumericConstant() throws TokenStreamException {
		int tokenIndex = 1;
		if (LA(tokenIndex) == PLUS || (LA(tokenIndex) == MINUS)) {
			if (LA(tokenIndex+1) == EOF || existsWhitespace(tokenIndex,tokenIndex+1)) {
				return false;
			}
			++tokenIndex;
		}
		switch (LA(tokenIndex)) {
			case GENERICOPTIMTOKEN: {
				// (decimalString)- (...) is error. (deciamlString)-(...) is not error
				if (LA(tokenIndex+1) == MINUS) {
					if (!existsWhitespace(tokenIndex,tokenIndex+1)) {
						if (LA(tokenIndex+2) != EOF && existsWhitespace(tokenIndex+1,tokenIndex+2)) {
							return false;
						}
					}
				}
				// (-)?(decimalString)(.(decimalString)?)?
				String tokenString = LT(tokenIndex).getText();
				return Character.isDigit(tokenString.charAt(0)) && (tokenString.length() < 2 
					|| Character.isDigit(tokenString.charAt(1)));
			}
			case DOT: {
				// (+|-).decimalString
				return LA(tokenIndex+1) == GENERICOPTIMTOKEN && !existsWhitespace(tokenIndex,tokenIndex+1)
					&& Character.isDigit(LT(tokenIndex+1).getText().charAt(0));
			}
		}
	    return false;
	}

	/**
	 * Determines if a hexadecimal literal appears in the token stream.
	 * @throws TokenStreamException anything that goes wrong while generating a stream of tokens 
	 */
	protected boolean possibleHexadecimalConstant() throws TokenStreamException {
		switch (LA(1)) {
			case GENERICOPTIMTOKEN:
				String tokenString = LT(1).getText();
				return tokenString.length() > 3 && tokenString.charAt(0) == '0' && "Xx".indexOf(tokenString.charAt(1)) >= 0;
			case LITERAL_X:
				return LA(2) == SINGLEQUOTE && LA(4) == SINGLEQUOTE && !existsWhitespace(1,4);
		}
		return false;
	}


	/**
	 * Append the given character of the given count 
	 * @param sb a string buffer
	 * @param ch a character
	 * @param count a repeat count
	 * @returns a string buffer
	 */
	protected StringBuilder appendChar(StringBuilder sb, char ch, int count) {
		for (int i=0; i < count; i++) {
			sb.append(ch);
		}
		return sb;
	}

	/**
	 * Set the column data type
	 * @param flavor the class of data type (string, numeric, binary, datetime, or boolean)
	 * @param length the column length (valid for the other than numeric)
	 * @param precision the precision of numeric type 
	 * @param scale the scale of numeric type
	 */
	public void setDataType(int flavor, int length, int precision, int scale, boolean nullable) {
		this.flavor = flavor;
		this.length = length;
		this.precision = precision;
		this.scale = scale;
		this.nullable = nullable;
	}
}
	
parse
	: columnexpression EOF 
	;

columnexpression
	: columnexpression1 (operator columnexpression2)*
	;

parseRelationship
	: relationshipexpression EOF
	;

relationshipexpression
	: (columnexpression1base | substring) (concatinate columexpression2base)*
	;

columnexpression1 { handleOracleSequenceLookAhead(); }	
    : specialRegisters
    | exitRoutine
	| procedure
	| generalfunctions
	| privacyfunctions
	| { flavor == datetimeFlavor }? datetimeLiteral
	| { possibleOracleSequence() }? oracleSequence
	| columnexpression1base
	;

columnexpression1base
	: nullConstant
    | booleanConstant
    | { flavor != datetimeFlavor }? stringLiteral
    | { possibleNumericConstant() }? numericConstant
    | { possibleHexadecimalConstant() }? hexadecimalLiteral
	| explicitColumnName
	;

columnexpression2
	: random
	| sequential
    | columexpression2base
	;

columexpression2base
	: stringLiteral
	| substring
    | { possibleNumericConstant() }? numericConstant
    | { possibleHexadecimalConstant() }? hexadecimalLiteral
	| explicitColumnName
	;

operator 
	: (	{ (flavor & numericFlavor) != 0 && numOperators == 0 && !possibleNumericConstant() || (flavor & binaryFlavor) != 0 }? PLUS
	  | { (flavor & numericFlavor) != 0 && numOperators == 0 && !possibleNumericConstant()}? (MINUS | STAR | SLASH)
	  | { (flavor & binaryFlavor) != 0 }? (BARS | "CONCAT")
	  | concatinate
	  )
	  { numOperators++; }
	;

concatinate
	: { (flavor & stringFlavor) != 0 }? (PLUS | BARS | "CONCAT")
	;

specialRegisters
	: { flavor == unknownFlavor || (flavor & datetimeFlavor) != 0 }?
	  ( "CURRENT" "DATE" | "CURRENT" "TIME" | "CURRENT" "TIMESTAMP"
	  | "CURRENT_DATE" | "CURRENT_TIME" | "CURRENT_TIMESTAMP"
	  | ("CURDATE" | "GETDATE" | "CURTIME" | "GETTIME") (noparameter)?
	  | "NOW" (noparameter)?
	  | "SYSDATE" (noparameter)?
	  )
	| { flavor == unknownFlavor || (flavor & stringFlavor) != 0 }?
	  ( "CURRENT_SQLID" | "CURRENT" "SQLID" | "USER" | "WORKSTATION_ID"
	  )
	;

generalfunctions
	: substring | random | sequential | identity | serial | propagate
	;

privacyfunctions
	: transSsn | transCcn | transEml | transfill | transCol | transNid | transAff | lookup | hash_lookup | rand_lookup | shuffle | age | currency
	;

identity
	: { flavor == unknownFlavor || (flavor & numericFlavor) != 0 }?
	  "IDENTITY" noparameter
	;

serial
	: { flavor == unknownFlavor || (flavor & numericFlavor) != 0 }?
	  "SERIAL" noparameter
	;

sequential { int start, step;}
	: { flavor == unknownFlavor || (flavor & (stringFlavor | numericFlavor)) != 0 }?
	  "SEQ" LEFTPAREN start=integer COMMA step=integer RIGHTPAREN
	  { step != 0 }?
	  { if (flavor == unknownFlavor) flavor = stringFlavor; }
	  { flavor != stringFlavor || String.valueOf(start).length() <= length && String.valueOf(step).length() <= length }?
	  { flavor != numericFlavor || String.valueOf(Math.abs(start)).length() <= precision - scale && String.valueOf(Math.abs(step)).length() <= precision - scale }? 
	;

random { int low, high; }
	: { flavor == unknownFlavor || (flavor & (stringFlavor | numericFlavor)) != 0 }?
	  "RAND" LEFTPAREN low=integer COMMA high=integer RIGHTPAREN
	  { low < high }?
	  { if (flavor == unknownFlavor) flavor = stringFlavor; }
	  { flavor != stringFlavor || String.valueOf(low).length() <= length }?
	  { flavor != numericFlavor || String.valueOf(Math.abs(low)).length() <= precision - scale }? 
	;

substring { int start=-1, length=-1; }
	: { flavor == unknownFlavor || (flavor & (stringFlavor | binaryFlavor)) != 0 }?
	  "SUBSTR" LEFTPAREN columnNameOrDoubleQuotedString COMMA start=unsignedInteger (COMMA length=unsignedInteger)? RIGHTPAREN
	  //TODO check if the specified column is compatible with the destination column
	  { start != 0 && length != 0 }?
	  { if ((flavor & (stringFlavor | binaryFlavor)) == 0) { flavor = (stringFlavor | binaryFlavor); } } 
	  //TODO should set actual data types of the column
	  //TODO should check start + length - 1 <= the length of specified column
	;

oracleSequence
	: { flavor == unknownFlavor || (flavor & numericFlavor) != 0 }? 
	  creatorID DOT sequenceName DOT "NEXTVAL" (LEFTPAREN "INCL_UPD" RIGHTPAREN)?
	;

propagate
	: { ++inPropFunc; }
	  "PROP" LEFTPAREN (
		columnexpression ((COMMA)? columnName)?
//	  | "EXIT" genericOptimToken
//	  | "PROC" ("LOCAL" | genericOptimToken DOT genericOptimToken)
	  ) RIGHTPAREN
	  { --inPropFunc; }
	;

exitRoutine
	: "EXIT" dllname
	    (LEFTPAREN ( (singleQuotedString|numericNumber) { int parmCnt = 1; } ({ ++parmCnt <= 8 }? COMMA (singleQuotedString|numericNumber))* )? RIGHTPAREN)?
	;

procedure
	: "PROC" ({ !existsWhitespace(1,3) }? objectFullName | "LOCAL")
		(LEFTPAREN ( (singleQuotedString|numericNumber) { int parmCnt = 1; } ({ ++parmCnt <= 8 }? COMMA (singleQuotedString|numericNumber))* )? RIGHTPAREN)?
	  //TODO should check if procedure exists 
	;

transSsn
	: "TRANS" "SSN" (LEFTPAREN SINGLEQUOTE 
		{ handlePossibleGenericOptimToken(new int[]{ EQUALS, DOUBLEQUOTEDSTRING, LITERAL_PRESERVE, SINGLEQUOTE, MINUS }); } 
	    ( transSsnFlags)?
		{ handlePossibleGenericOptimToken(new int[]{ DOUBLEQUOTEDSTRING, LITERAL_PRESERVE, SINGLEQUOTE, MINUS }); } 
		(columnName | DOUBLEQUOTEDSTRING)? 
		("PRESERVE" EQUALS "INVALID")? 
	  SINGLEQUOTE RIGHTPAREN)?
	;

transSsnFlags
	: { String flags = "";
		int nextCharPos = LT(1).getColumn() + 1; }
	  EQUALS 
	  ( { LT(1).getColumn() == nextCharPos }?
	    { flags += LT(1).getText();
	      nextCharPos += LT(1).getText().length(); }
	    ~(SINGLEQUOTE|RIGHTPAREN|EOF)
	  )+
	  { isValidFlags(flags, "NMRVI-") && countValidChars(flags,"NR") != 2 }?
	;

transCcn
	: "TRANS" "CCN" (LEFTPAREN SINGLEQUOTE 
		{ handlePossibleGenericOptimToken(new int[]{ EQUALS, DOUBLEQUOTEDSTRING, LITERAL_PRESERVE, SINGLEQUOTE }); } 
		( transCcnFlags)? 
		{ handlePossibleGenericOptimToken(new int[]{ DOUBLEQUOTEDSTRING, LITERAL_PRESERVE, SINGLEQUOTE }); } 
		( columnName | DOUBLEQUOTEDSTRING )? 
		( "PRESERVE" EQUALS "INVALID" )? 
	  SINGLEQUOTE RIGHTPAREN)?
	;

transCcnFlags
	: EQUALS
	  { isValidFlags(LT(1), "NR6I") && !containsConflictFlags(LT(1),"NR6") }?
	  genericOptimToken
	;

transEml
	: "TRANS" "EML" (LEFTPAREN SINGLEQUOTE 
		{ if (LA(1) != EQUALS) handleTransEmlLookAhead(); }
		(transEmlFlags)? 
		transEmlParameters 
	  SINGLEQUOTE RIGHTPAREN)?  
	;

transEmlFlags
	: { String flags = "";
		int nextCharPos = LT(1).getColumn() + 1; }
	  EQUALS 
	  ( { LT(1).getColumn() == nextCharPos }?
	    { flags += LT(1).getText();
	      nextCharPos += LT(1).getText().length(); }
	    ~(COMMA|SINGLEQUOTE|RIGHTPAREN|EOF)
	  )+
	  { isValidFlags(flags, "N._ILUE") }?
	;

transEmlParameters {
 	  int commaCnt = 0;
	  boolean sourceColSet = false;
	  boolean prefixSet = false;
	  boolean preserveSet = false;
	  handleTransEmlLookAhead();
	}
	: ( COMMA { commaCnt++; } )? 
  	  { handleTransEmlLookAhead(); }
	  ( columnNameOrDoubleQuotedString { sourceColSet = true; } )?
  	  { handleTransEmlLookAhead(); }
  	  ( (COMMA { commaCnt++; })? 
  	  { handleTransEmlLookAhead(); }
	  ( { !prefixSet && !preserveSet && (sourceColSet || commaCnt == 2) }? transEmlPrefix { prefixSet = true; } 
	  | { !preserveSet }? "PRESERVE" EQUALS "INVALID" { preserveSet = true; } ) )* 
	;

transEmlPrefix { handlePossibleGenericOptimToken(new int[]{ DOUBLEQUOTEDSTRING }); }
	: columnName { handlePossibleGenericOptimToken(new int[]{ COMMA, SINGLEQUOTE }); } (columnName)? 
	| DOUBLEQUOTEDSTRING
	;

transfill {
	  boolean repcharSet = false;
	  boolean lengthSet = false;
	}
	: "TRANS" "FILL" LEFTPAREN SINGLEQUOTE
		(COMMA
		| "REPCHAR" EQUALS oneCharacter { repcharSet = true; }
		| "LENGTH" EQUALS (unsignedInteger|"MATCH"|"MAX") { lengthSet = true; }
		)+
	  SINGLEQUOTE RIGHTPAREN
	  { repcharSet && lengthSet }?
	;

transCol
	: "TRANS" "COL" LEFTPAREN SINGLEQUOTE ( ("UNIQUE" | "HASH")
		{ handleTransColLookAhead(); }
		((COMMA)? "SOURCE" EQUALS columnNameOrDoubleQuotedString)?
		((COMMA)? "COPY" EQUALS LEFTPAREN (LEFTPAREN unsignedInteger COMMA unsignedInteger (COMMA DOUBLEQUOTEDSTRING)? RIGHTPAREN)+ RIGHTPAREN)?
		((COMMA)? "SEED" EQUALS (DOUBLEQUOTEDSTRING | "VAR" LEFTPAREN genericOptimToken RIGHTPAREN | "RANDOM"))?
		((COMMA)? "LENGTH" EQUALS (unsignedInteger | "MAX"))?
		((COMMA)? "PRESERVE" EQUALS LEFTPAREN (ignoreChoice | (COMMA ignoreChoice))+ RIGHTPAREN)?
		((COMMA)? "TRIM" EQUALS LEFTPAREN oneCharacter (COMMA oneCharacter)* (BACKSLASH "U" (BACKSLASH "R")? | BACKSLASH "R" (BACKSLASH "U")?)? RIGHTPAREN)?
		((COMMA)? "NUM")?
	  ) SINGLEQUOTE RIGHTPAREN
	;

transNid
	: ("TRANS")? ("NID"|"NATIONAL_ID") LEFTPAREN SINGLEQUOTE ("SWI" | "SWITCH") EQUALS countryCode
		(COMMA ((("FMT" | "FORMAT") EQUALS LEFTPAREN outputFormat RIGHTPAREN) |
		        (("MTD" | "METHOD") EQUALS ("MASK" | "RANDOM")) |
		        (("SRC" | "SOURCE") EQUALS columnNameOrDoubleQuotedString) |
		        (("VAL" | "VALIDATE") EQUALS ("Y" | "N")) |
		        (("PRE" | "PRESERVE") EQUALS ("INV" | "INVALID"))
		       )?
		)*
	  SINGLEQUOTE RIGHTPAREN   
	;

transAff
    : { boolean fieldDefined = false;
        boolean keyDefined = false;
        boolean algoDefined = false;
        int algo = LITERAL_DEFAULT;
        boolean iterationsDefined = false;
        boolean tweaksDefined = false;
		boolean methodDefined = false;
		int method = UNKNOWN;
		boolean lengthDefined = false;
		boolean seedDefined = false;
		boolean copyDefined = false;
		int copyStart = 0;
		String copyLiteral = null;
		boolean ruleDefined = false;
		int rule = UNKNOWN;
		boolean caseDefined = false;
		boolean removeDefined = false;
		boolean trimDefined = false;
      }
	  "TRANS" ("PRO" | "PROVIDER") EQUALS ("AFF" | "AFFINITY") 
        ( COMMA (
          { LT(1).getText().matches("(?i)FLDDEF\\d\\d?") }? genericOptimToken EQUALS LEFTPAREN
		    "NAME" EQUALS columnNameOrDoubleQuotedString COMMA ("DT" | "DATATYPE") EQUALS affDataType
            ( COMMA (
              ("PRE" | "PRECISION") EQUALS unsignedInteger
            | ("SCA" | "SCALE") EQUALS unsignedInteger
            | ("LEN" | "LENGTH") EQUALS unsignedInteger
            | ("CP" | "CODEPAGE") EQUALS affCodePage
            | ("CPT" | "CPTYPE") EQUALS affCodePage
            ) )*
            RIGHTPAREN
            { fieldDefined = true; }
        | ("ALGO" | "ALGORITHM") EQUALS { algo = LA(1); } ("DEF" | "DEFAULT" | "FPE") { algoDefined = true; }
        | "KEY" EQUALS ({ LT(1).getText().length() <= 34 }? DOUBLEQUOTEDSTRING | { LT(1).getText().startsWith("@") }? genericOptimToken) { keyDefined = true; }
        | "ITERATIONS" EQUALS { between(LT(1),1,99) }? unsignedInteger { iterationsDefined = true; }
        | "TWEAKS" EQUALS ("YES" | "NO") { tweaksDefined = true; }
        | "WHENMATCH" EQUALS (
              "REVERSE" 
            | "SHUFFLE" 
            | "PREFIX" LEFTPAREN ( DOUBLEQUOTEDSTRING | "SEQ" LEFTPAREN { int low, high; } low=unsignedInteger COMMA high=unsignedInteger { low <= high && high <= 2000000000 }? RIGHTPAREN | "RAND" LEFTPAREN { int low, high; } low=unsignedInteger COMMA high=unsignedInteger { low <= high && high <= 2000000000 }? RIGHTPAREN) RIGHTPAREN
            | "SUFFIX" LEFTPAREN ( DOUBLEQUOTEDSTRING | "SEQ" LEFTPAREN { int low, high; } low=unsignedInteger COMMA high=unsignedInteger { low <= high && high <= 2000000000 }? RIGHTPAREN | "RAND" LEFTPAREN { int low, high; } low=unsignedInteger COMMA high=unsignedInteger { low <= high && high <= 2000000000 }? RIGHTPAREN) RIGHTPAREN
            | "OVERLAY" LEFTPAREN unsignedInteger COMMA unsignedInteger COMMA DOUBLEQUOTEDSTRING RIGHTPAREN 
            )
        | ("METHOD" | "MTD") EQUALS { method = LA(1); } ("REP" | "REPEATABLE" | "HASH") { methodDefined = true; }
        | ("LENGTH" | "LEN") EQUALS (unsignedInteger | "MAX") { lengthDefined = true; }
        | "SEED" EQUALS (DOUBLEQUOTEDSTRING | { LT(1).getText().startsWith("@") }? genericOptimToken | "RAN" | "RANDOM") { seedDefined = true; }
        | "COPY" EQUALS (LEFTPAREN copyStart=unsignedInteger COMMA unsignedInteger (COMMA { copyLiteral = LT(1).getText(); } DOUBLEQUOTEDSTRING)? RIGHTPAREN)+ { !( copyLiteral != null && copyStart >= copyLiteral.length() - 2 )}? { copyDefined = true; }
        | "RULE" EQUALS { rule = LA(1); } ("NUM" | "NUMERIC" | "BIN" | "BINARY" | "SCA" | "SCALED") { ruleDefined = true; }
        | "CASE" EQUALS ("UP" | "UPPER") { caseDefined = true; }
        | "REMOVE" EQUALS DOUBLEQUOTEDSTRING { removeDefined = true; }
        | "TRIM" EQUALS "RIGHT" { trimDefined = true; }
        | ("DLIM" | "DISCARDLIMIT") EQUALS unsignedInteger
        | ("CP" | "CODEPAGE") EQUALS affCodePage
        | ("CPT" | "CPTYPE") EQUALS affCodePageType
        ) )*
      { fieldDefined && methodDefined }?
      { !( (algo == LITERAL_DEF || algo == LITERAL_DEFAULT) && (keyDefined || iterationsDefined || tweaksDefined) ) }?
      { !( algo == LITERAL_FPE && (seedDefined || method == LITERAL_HASH) ) }?
      { !( ruleDefined && (copyDefined || lengthDefined) ) }?
      { !( (rule == LITERAL_BIN || rule == LITERAL_BINARY) && (copyDefined || lengthDefined || trimDefined || removeDefined || caseDefined) ) }?
      { !( lengthDefined && (algo == LITERAL_DEF || algo == LITERAL_DEFAULT) != (method == LITERAL_HASH) ) }?
      { !( (caseDefined || removeDefined || trimDefined) && (rule == LITERAL_BIN || rule == LITERAL_BINARY || rule == LITERAL_SCALED) ) }?
    ;

affDataType
	: "CHAR"
	| "DATE"
	| "DATETIME_CHAR"
	| "DATETIME_SZ"
	| "DATETIME_VARCHAR"
	| "DATETIME_WCHAR"
	| "DATETIME_WSZ"
	| "DATETIME_WVARCHAR"
	| "DECIMAL_370"
	| "DOUBLE"
	| "FLOAT"
	| "INTEGER"
	| "LONG_LONG"
	| "ORA_VARNUM"
	| "SMALLINT"
	| "TIME"
	| "TIMESTAMP"
	| "TINYINT"
	| "U_INTEGER"
	| "U_LONG_LONG"
	| "U_SMALLINT"
	| "U_TINYINT"
	| "VARCHAR"
	| "VARCHAR_SZ"
	| "WCHAR"
	| "WVARCHAR_SZ"
	;

affCodePage 
    : genericOptimToken (MINUS genericOptimToken)?
	;

affCodePageType
	: "DBZ" | "DB2ZOS" 
	| "DB2" | "DB2LUW" 
	| "ORA" | "ORACLE" 
	| "SYB" | "SYBASE" 
	| "ODBC" 
	| "IFX" | "INFORMIX" 
	| "MSS" | "SQLSERVER" 
	| "TD"  | "TERADATA" 
	| "NZ"  | "NETEZZA" 
	| "ANY" 
	| "NONE"
	;

lookup
	: "LOOKUP" LEFTPAREN lookupParameters[false] RIGHTPAREN
	;

hash_lookup
	: "HASH_LOOKUP" LEFTPAREN lookupParameters[true] RIGHTPAREN
	;

lookupParameters[boolean isHashLookup]
	: { String destName = null;
		int numDestNames = 0;
		int numSrcNames = 0;
		int numIgnoreNames = 0;
		if (LA(2) != EQUALS) { handlePossibleGenericOptimToken(null); } }
	  ( columnName COMMA | "SRCSEARCH" EQUALS LEFTPAREN columnList RIGHTPAREN COMMA )? 
	  { if (LA(2) != EQUALS) { handlePossibleGenericOptimToken(null); } }
	  ( { isHashLookup }? "TRIM" EQUALS LEFTPAREN ({ LA(1) != BACKSLASH || LA(2) != LITERAL_U }? trimChar)* (BACKSLASH "U")? RIGHTPAREN COMMA )? 
	  { if (LA(2) != EQUALS) { handlePossibleGenericOptimToken(null); } }
      ( "DEST" EQUALS LEFTPAREN numDestNames = columnList RIGHTPAREN COMMA)? 
	  { handlePossibleGenericOptimToken(new int[]{ COMMA, RIGHTPAREN }); }
	  qualifiedTableName LEFTPAREN 
		{ if (LA(2) != EQUALS) { handlePossibleGenericOptimToken(null);
			if (LA(4) != EQUALS) { handlePossibleGenericOptimToken(3, null); } } } 
		( columnName | "LKPSEARCH" EQUALS LEFTPAREN columnList RIGHTPAREN ) COMMA 
			{ if (LA(2) != EQUALS) { handlePossibleGenericOptimToken(null); } } 
			( columnName | ("VALUES" | "SOURCE") EQUALS LEFTPAREN numSrcNames = columnList RIGHTPAREN ) 
		RIGHTPAREN
	  {
		boolean cacheSet = false;
		boolean ignoreSet = false;
		boolean seedSet = false;
	  }  
	  ( COMMA (
		  { !cacheSet }? ("CACHE" | "NOCACHE") { cacheSet = true; }
		| {	!ignoreSet }? { int ignoreOrPreserve = LA(1); } ("IGNORE" | "PRESERVE") EQUALS LEFTPAREN numIgnoreNames = ignoreColList[ignoreOrPreserve == LITERAL_PRESERVE] RIGHTPAREN { ignoreSet = true; }
		| { !seedSet }? { isHashLookup}? "SEED" EQUALS { between(LT(1), 0, 2000000000) }? unsignedInteger { seedSet = true; }
	  ) )*
	  { numDestNames == numSrcNames }?
	  { numIgnoreNames == 0 || numDestNames == 0 && numIgnoreNames == 1 || numDestNames > 0 && numDestNames == numIgnoreNames }?
	;

trimChar
	: BACKSLASH (RIGHTPAREN | BACKSLASH) 
	| ~(BACKSLASH | RIGHTPAREN)
	;

rand_lookup {
	  int numDestNames = 0;
	  int numSrcNames = 0;
	  int numIgnoreNames = 0;
	}
	: "RAND_LOOKUP" LEFTPAREN 
			qualifiedTableName COMMA 
			{ if (LA(2) != EQUALS) { handlePossibleGenericOptimToken(null); } } 
			(columnName | "DEST" EQUALS LEFTPAREN numDestNames = columnList RIGHTPAREN COMMA ("SOURCE" | "VALUES") EQUALS LEFTPAREN numSrcNames = columnList RIGHTPAREN)
			(COMMA { between(LT(1), 0, 2000000000) }? unsignedInteger)?
			(COMMA ("IGNORE" | "PRESERVE" ) EQUALS LEFTPAREN numIgnoreNames = ignoreColList[false] RIGHTPAREN)?
		RIGHTPAREN 
	  { numDestNames == numSrcNames }?
	  { numIgnoreNames == 0 || numDestNames == 0 && numIgnoreNames == 1 || numDestNames > 0 && numDestNames == numIgnoreNames }?
	;

shuffle {
	  boolean destSet = false;
	  boolean retrySet = false;
	  boolean ignoreSet = false;
	  int numDestNames = 0;
	  int numIgnoreNames = 0;
	}
	: "SHUFFLE" (LEFTPAREN 
	    { boolean commaRequired = false; }
	    ( ({ commaRequired }? COMMA { commaRequired = false;})? ({ !commaRequired }? ( 
		    { !destSet }? "DEST" EQUALS LEFTPAREN numDestNames = columnList RIGHTPAREN { destSet = true; }
		  | { !retrySet }? "RETRY" (EQUALS { between(LT(1), 0, 1000) }? unsignedInteger)? { retrySet = true; }
	      | { !ignoreSet }? "IGNORE" EQUALS LEFTPAREN numIgnoreNames = ignoreColList[false] RIGHTPAREN { ignoreSet = true; }
	    ) { commaRequired = true; } ) )+
	  RIGHTPAREN)? 
	  { numIgnoreNames == 0 || numDestNames == 0 && numIgnoreNames == 1 || numDestNames > 0 && numDestNames == numIgnoreNames }?
	  { inPropFunc == 0 || numDestNames == 0 }?
	;

columnList returns[int numColumnNames = 0]
	: { ++numColumnNames; } columnName (COMMA { ++numColumnNames; } columnName)* 
	;

ignoreColList[boolean allowNotFound] returns[int numIgnoreNames = 0]
	: { boolean commaRequired = false; 
		boolean notFoundSet = false; }
	  { if (LA(2) == LEFTPAREN) handlePossibleGenericOptimToken(null); } 
	  ( { allowNotFound }? "NOT_FOUND" { commaRequired = true; notFoundSet = true; })?
	  ( ( { commaRequired }? COMMA {commaRequired = false;})? { !commaRequired }?
		{ ++numIgnoreNames; } columnName LEFTPAREN (ignoreChoice (COMMA ignoreChoice)*)? RIGHTPAREN 
		{ commaRequired = true; } )* 
	  { notFoundSet || numIgnoreNames > 0 }?
	;

ignoreChoice
	: "SPACES" | "NULL" | "ZERO_LEN"
	;

countryCode
	: "CA" | "FR" | "IT" | "ES" | "UK" | "US"
	;

outputFormat
	: countryCode EQUALS (genericOptimToken|DOT|MINUS)+
	;

age
	{ boolean noneFound = false; }
	: { flavor == unknownFlavor || (flavor & (datetimeFlavor | stringFlavor)) != 0 || (flavor & numericFlavor) != 0 && scale == 0 }?
      "AGE" LEFTPAREN ("NONE" { noneFound = true; } | ageParameters) RIGHTPAREN
	  { !((flavor & ~datetimeFlavor) != 0 && noneFound) }?
	;

ageParameters {
	  boolean defFound = false;
	  boolean srcColFound = false;
	  boolean numIncrFound = false;
	  boolean yearIncrFound = false;
	  boolean fixedYearFound = false;
	  boolean monthIncrFound = false;
	  boolean weekIncrFound = false;
	  boolean dayIncrFound = false;
	  boolean ruleIncrFound = false;
	  boolean calendarFound = false;
	  boolean ruleFound = false;
	  boolean pivotFound = false;
	  boolean srcFmtFound = false;
	  boolean dstFmtFound = false;
	  boolean srcExitFound = false;
	  boolean dstExitFound = false;
	}
	: ( COMMA
	  | { !(numIncrFound || defFound || fixedYearFound || ruleIncrFound) }? "DEF" { defFound = true; }
	  | { !srcColFound }? ("SC" | "SRCCOL") EQUALS columnName { srcColFound = true; }
	  | { boolean plus = false; boolean minus = false; }
	    (PLUS { plus = true; } | MINUS { minus = true; })? { String tokenText = LT(1).getText(); } agingAmount[plus, minus] 
	    { char lastChar = tokenText.charAt(tokenText.length() - 1); }
	    ( { lastChar == 'Y' }? { int value = Integer.valueOf(tokenText.substring(0, tokenText.length() - 1)); }
	      ( { 1582 <= value && value <= 3999}? () { !(numIncrFound || defFound || fixedYearFound || ruleIncrFound)}? { fixedYearFound = true; }  
	      | { -2500 <= value && value <= 1581}? () { !(yearIncrFound || defFound || fixedYearFound || ruleIncrFound) }? { yearIncrFound = true; numIncrFound = true; }
	      )
	    | { lastChar == 'M' }? () { !(monthIncrFound || defFound || fixedYearFound || ruleIncrFound) }? { monthIncrFound = true; numIncrFound = true; }
	    | { lastChar == 'W' }? () { !(weekIncrFound || defFound || fixedYearFound || ruleIncrFound) }? { weekIncrFound = true; numIncrFound = true; }
	    | { lastChar == 'D' }? () { !(dayIncrFound || defFound || fixedYearFound || ruleIncrFound)  }? { dayIncrFound = true; numIncrFound = true; }
	    | { lastChar == 'R' }? () { !(numIncrFound || defFound || fixedYearFound || ruleIncrFound) }? { ruleIncrFound = true; }
	    )
	  | { !calendarFound }? ("CA" | "CALENDAR") EQUALS calendarName { calendarFound = true; }
	  | { !ruleFound }? ("RU" | "RULE") EQUALS (ruleName | "DEF") { ruleFound = true; }
	  | { !pivotFound }? ("PI" | "PIVOT") EQUALS { LT(1).getText().length() == 2 && between(LT(1), 0, 99) }? unsignedInteger { pivotFound = true; }
	  | { !(srcFmtFound || srcExitFound) }? ("SF" | "SRCFMT") EQUALS dateFormat { srcFmtFound = true; }
	  | { !(dstFmtFound || dstExitFound) }? ("DF" | "DSTFMT") EQUALS dateFormat { dstFmtFound = true; }
	  | { !(srcFmtFound || srcExitFound) }? ("SX" | "SRCEXIT") EQUALS dllname { srcExitFound = true; }
	  | { !(dstFmtFound || dstExitFound) }? ("DX" | "DSTEXIT") EQUALS dllname { dstExitFound = true; }
	  )+
	  { (defFound ? 1 : 0) + (numIncrFound ? 1 : 0) + (fixedYearFound ? 1 : 0) + (ruleIncrFound ? 1 : 0) == 1 }?
	  { !(ruleIncrFound && !ruleFound) }?
	  { !(calendarFound && !ruleFound) }?
	  { !(!srcColFound && (flavor & ~datetimeFlavor) != 0 && (defFound || !srcFmtFound)) }?
	;

agingAmount[boolean plus, boolean minus] 	  
	: { isValidAgingAmount(LT(1), plus, minus) }? genericOptimToken
	;

calendarName
	: //TODO check if valid calendar name
	  genericOptimToken
	;

ruleName
	: //TODO check if valid rule name
	  genericOptimToken
	;

dateFormat
	{ String value; }
	: value = singleQuotedString
	  { isValidDateFormatString(value) }?
	;

genericFilename
	: genericOptimToken (DOT (genericOptimToken)?)?
	;

dllname
	: { String value = "";
		int nextCharPos = LT(1).getColumn(); }
	  ( { LT(1).getColumn() == nextCharPos }?
	  	{ value += LT(1).getText(); 
	  	  nextCharPos += LT(1).getText().length(); }
	    ~(COMMA|SINGLEQUOTE|LEFTPAREN|RIGHTPAREN)
	  )+
	  { isValidDllName(value) }?
	;

currency {
	  boolean srctypSet = false;
	  boolean srcspcSet = false;
	  boolean dsttypSet = false;
	  boolean dstspcSet = false;
	  boolean srccolSet = false;
	  boolean triangSet = false;
	  boolean curtblSet = false;
	  boolean trndatSet = false;
	  boolean datfmtSet = false;
	  boolean numsclSet = false;
	}
	: { flavor == unknownFlavor || (flavor & numericFlavor) != 0 }?
	  "CURRENCY" LEFTPAREN ( COMMA
		| { !srctypSet && !srcspcSet }? ("ST" | "SRCTYP") EQUALS currencyCode { srctypSet = true; }
		| { !srctypSet && !srcspcSet }? ("SS" | "SRCSPC") EQUALS LEFTPAREN columnName COMMA lookupTableNumber RIGHTPAREN { srcspcSet = true; }
		| { !dsttypSet && !dstspcSet }? ("DT" | "DSTTYP") EQUALS currencyCode { dsttypSet = true; }
		| { !dsttypSet && !dstspcSet }? ("DS" | "DSTSPC") EQUALS LEFTPAREN columnName COMMA lookupTableNumber RIGHTPAREN { dstspcSet = true; }
		| { !srccolSet }? ("SC" | "SRCCOL") EQUALS columnName { srccolSet = true; }
		| { !triangSet }? ("TR" | "TRIANG") { triangSet = true; }
		| { !curtblSet }? ("CU" | "CURTBL") EQUALS currencyDefinitionName { curtblSet = true; } 
		| { !trndatSet }? ("TD" | "TRNDAT") EQUALS columnName { trndatSet = true; }
		| { !datfmtSet }? ("DF" | "DATFMT") EQUALS dateFormat { datfmtSet = true; }
		| { !numsclSet }? ("NS" | "NUMSCL") EQUALS oracleNumericScale { numsclSet = true; }
	  )+ RIGHTPAREN
	  { (srctypSet || srcspcSet) && (dsttypSet || dstspcSet) }?
	;

currencyCode
	: { LT(1).getText().length() <= 3 }? genericOptimToken
	;

lookupTableNumber 
	: { between(LT(1), 1, 4) }? unsignedInteger
	;

currencyDefinitionName
	: genericOptimToken
	;

oracleNumericScale 
	: { between(LT(1), 0, 8) }? unsignedInteger
	;

nullConstant
	: { nullable }?
	  "NULL"
	;

numericConstant 
    { String values[]; }
	: { flavor == unknownFlavor || (flavor & numericFlavor) != 0 }?
	  values = numericNumber
	  { numOperators > 0 || values[0].length() + values[1].length() <= precision && values[1].length() <= scale }?
	  { flavor = numericFlavor; }
	;

booleanConstant
	: { flavor == unknownFlavor || (flavor & booleanFlavor) != 0 }?
	  "TRUE" | "FALSE"
	;

/*
 * A hexadecimal literal has one of these formats:
 * 		X'1234567890ABCDEF' or 0x1234567890ABCDEF  
 */
hexadecimalLiteral
	{ String hexstring; }
	: { flavor == unknownFlavor || ((flavor & binaryFlavor) != 0) || ((flavor & stringFlavor) != 0) }? 
	  ( { isValidHexString(hexstring = LT(3).getText()) && !existsWhitespace(1,4) }? "X" SINGLEQUOTE genericOptimToken SINGLEQUOTE
	  | { LT(1).getText().toUpperCase().startsWith("0X") && isValidHexString(hexstring = LT(1).getText().substring(2)) }? genericOptimToken
	  )
	  { hexstring.length() % 2 == 0 && hexstring.length() / 2 <= length }?
	  { flavor = binaryFlavor; }
	;

stringLiteral
	{ String value; }
	: { flavor == unknownFlavor || ((flavor & binaryFlavor) != 0) || ((flavor & stringFlavor) != 0) }?
	  value = singleQuotedString
	  { value.length() <= length }?
	  { flavor = stringFlavor; }
	;

datetimeLiteral
	{ String value; }
	: { flavor == unknownFlavor || (flavor & datetimeFlavor) != 0 }?
	  value = singleQuotedString
	  { isValidDateValueString(value) }? 
	;

explicitColumnName
	: columnNameOrDoubleQuotedString
	  { if (flavor == unknownFlavor) flavor = (stringFlavor | numericFlavor | binaryFlavor); }
	  //TODO should set actual data type of the column 
	;

noparameter
	: LEFTPAREN RIGHTPAREN
	;

decimalString returns[String value = LT(1).getText()]
	: { isValidDecimalString(LT(1)) }? genericOptimToken
	;

minusOrDecimalString returns[String value = LT(1).getText()]
	: { LT(1).getText().charAt(0) == '-' && isValidDecimalString(LT(1).getText().substring(1)) }?  genericOptimToken
	| decimalString
	;

integer returns[int value]
	: { isValidInteger(LT(1)) }? { value = Integer.valueOf(LT(1).getText()); } genericOptimToken 
	| { !existsWhitespace(1,2) && isValidUnsignedInteger(LT(2)) }? { value = Integer.valueOf(LT(2).getText()); } PLUS genericOptimToken
	| { !existsWhitespace(1,2) && isValidInteger(LT(1).getText()+LT(2).getText()) }? { value = Integer.valueOf(LT(1).getText()+LT(2).getText()); } MINUS genericOptimToken
	;

unsignedInteger returns[int value]
	: { isValidUnsignedInteger(LT(1)) }? { value = Integer.valueOf(LT(1).getText()); } genericOptimToken
	;

numericNumber returns[String[] values] 
	{ String integerPart = "", fractionPart = ""; }
	: { int numTokens = countClingingTokens(); }
	  ( { numTokens >= 3 && "+-".indexOf(LT(1).getText()) >= 0 && LA(2) == DOT }? (PLUS|MINUS) DOT fractionPart=decimalString
	  | { numTokens >= 2 }? DOT fractionPart=decimalString
	  | { numTokens >= 2 }? (PLUS|MINUS) integerPart=decimalString ({ numTokens >= 3 }? DOT ({ numTokens >= 4 }? fractionPart=decimalString)?)?
	  | integerPart=minusOrDecimalString ({ numTokens >= 2 }? DOT ({numTokens >= 3}? fractionPart=decimalString)?)?
	  )
	  { values = new String[]{ integerPart, fractionPart }; }
	;

singleQuotedString returns[String string]
	{ StringBuilder sb = new StringBuilder();
		int nextCharPos = LT(1).getColumn() + LT(1).getText().length(); }
	: SINGLEQUOTE ({
		  appendChar(sb, ' ', LT(1).getColumn() - nextCharPos).append(LT(1).getText());
		  nextCharPos = LT(1).getColumn() + LT(1).getText().length(); 
		} ~SINGLEQUOTE
	  )* 
	  { appendChar(sb, ' ', LT(1).getColumn() - nextCharPos); } SINGLEQUOTE
	  { string = sb.toString(); }
	;

oneCharacter
	: { LT(1).getText().length() == 1 }?
	  .
	;

qualifiedTableName { handleMultiPartNameLookAhead(); } 
	: ((dbAliasName DOT)? creatorID DOT)? tableName
	;

dbAliasName
	: //TODO check if dbAlias exists
	  optimIdentifier
	;

objectFullName
	: objectID DOT objectName
	;

objectID
	: { isValidObjectIdentifier(LT(1), 8) }?
	  .
	;

objectName
	: { isValidObjectIdentifier(LT(1), 12) }?
	  optimIdentifier
	;

creatorID
	: //TODO check if createID exists
	  optimIdentifier
	;

tableName
	: //TODO check if table exists
	  databaseObjectName
	;

columnName
	: //TODO check if column exists
	  databaseObjectName
	;

columnNameOrDoubleQuotedString { handlePossibleGenericOptimToken(new int[]{ DOUBLEQUOTEDSTRING }); }
	: columnName 
	| DOUBLEQUOTEDSTRING
	;

sequenceName
	: databaseObjectName
	;

optimIdentifier 
	: { isValidOptimIndentifierToken(LT(1)) }? genericOptimToken
	;

databaseObjectName
	: optimIdentifier
	;

genericOptimToken { handlePossibleGenericOptimToken(null); }
	: GENERICOPTIMTOKEN
	;

{ @SuppressWarnings("all") }
class DistributedColumnExpressionLexer extends Lexer;

options {
	k = 5;
	charVocabulary = '\u0000'..'\uFFFE';
	caseSensitiveLiterals = false;
}


GENERICOPTIMTOKEN
	: (
		~(
			' ' | 
			'\t'| 
			'\u001a' |
			'\n' |
			'\r' |
			'%' |
			'~' |
			'`' |
			'!' |
			'^' |
			'&' |
			'*' |
			'(' |
			')' |
			'+' |
			'-' |
			'=' |
			'|' |
			'{' |
			'}' |
			'[' |
			']' |
			':' |
			';' |
			'\"' |
			'\'' |
			'<' |
			'>' |
			',' |
			'.' |
			'?' |
			'/' |
			'\\'			
	 	)
	)+
	;

DOUBLEQUOTEDSTRING
	: DOUBLEQUOTE (~'"')+ DOUBLEQUOTE
	;

BACKSLASH
	: '\\'
	;

BARS
	: "||"
	;

COLON
	: ':'
	;

COMMA
	: ','
	;

DOT
	: '.'
	;

EQUALS
	: '='
	;

LEFTPAREN
	: '('
	;

MINUS
	: '-'
	;

PLUS
	: '+'
	;

RIGHTPAREN
	: ')'
	;

DOUBLEQUOTE
	: '"'
	;

SINGLEQUOTE
	: '\''
	;

SLASH
	: '/'
	;

STAR
	: '*'
	;

NEWLINE
	: ("\r\n" | '\n' | '\r') { $setType(Token.SKIP); newline();}
	;
WS
	: (' ' | '\t' | '\u001a') { $setType(Token.SKIP);}
	;

UNKNOWN
	: . 
	;
